# Copyright (c) 2014, Salesforce.com, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# - Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# - Neither the name of Salesforce.com nor the names of its contributors
#   may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''
A parameter-free clustering prior based on partition entropy.

The Dirichlet prior is often used in nonparametric mixture models to model
the partitioning of observations into clusters.
This class implements a parameter-free alternative to the Dirichlet prior
that preserves exchangeability, preserves asymptotic convergence rate of
density estimation, and has an elegant interpretation as a
minimum-description-length prior.

Motivation
----------

In conjugate mixture models, samples from the posterior are sufficiently
represented by a partitioning of observations into clusters, say as an
assignment vector X with cluster labels X_i for each observation i.
In our ~10^6-observation production system, we found the data size of this
assignment vector to be a limiting factor in query latency, even after
lossless compression.  To address this problem, we tried to incorporate the
Shannon entropy of this assignment vector directly into the prior, as
an information criterion.  Surprisingly, the resulting low-entropy prior
enjoys a number of properties:

- The low-entropy prior enjoys similar asymptotic convergence as the
  Dirichlet prior.

- The probability of a clustering is elegant and easy to evaluate
  (up to an unknown normalizing constant).

- The resulting distribution resembles a CRP distribution with parameter
  alpha = exp(-1), but slightly avoids small clusters.

- MAP estimates are minimum-description-length, as measured by assignment
  vector complexity.

- The low-entropy prior is parameter free, unlike the CRP, Pitman-Yor, or
  Mixture of Finite Mixture models.

A difficulty is that the prior depends on dataset size, and is hence not a
proper nonparametric generative model.
'''

import os
from collections import defaultdict
import numpy
from numpy import log, exp
import math
import matplotlib
matplotlib.use('Agg')
from matplotlib import pyplot, font_manager
import parsable
from distributions.lp.special import fast_log
from distributions.io.stream import json_stream_load, json_stream_dump
parsable = parsable.Parsable()

assert exp  # pacify pyflakes


DEFAULT_MAX_SIZE = 47

ROOT = os.path.dirname(os.path.abspath(__file__))
TEMP = os.path.join(ROOT, 'clustering.data')
if not os.path.exists(TEMP):
    os.makedirs(TEMP)

CACHE = {}


def savefig(stem):
    for extension in ['png', 'pdf']:
        name = '{}/{}.{}'.format(TEMP, stem, extension)
        print 'saving', name
        pyplot.tight_layout()
        pyplot.savefig(name)


def get_larger_counts(small_counts):
    large_counts = defaultdict(lambda: 0.0)
    for small_shape, count in small_counts.iteritems():

        # create a new partition
        large_shape = (1,) + small_shape
        large_counts[large_shape] += count

        # add to each existing partition
        for i in xrange(len(small_shape)):
            large_shape = list(small_shape)
            large_shape[i] += 1
            large_shape.sort()
            large_shape = tuple(large_shape)
            large_counts[large_shape] += count

    return dict(large_counts)


def get_smaller_probs(small_counts, large_counts, large_probs):
    assert len(large_counts) == len(large_probs)
    small_probs = {}
    for small_shape, count in small_counts.iteritems():
        prob = 0.0

        # create a new partition
        large_shape = (1,) + small_shape
        prob += large_probs[large_shape] / large_counts[large_shape]

        # add to each existing partition
        for i in xrange(len(small_shape)):
            large_shape = list(small_shape)
            large_shape[i] += 1
            large_shape.sort()
            large_shape = tuple(large_shape)
            prob += large_probs[large_shape] / large_counts[large_shape]

        small_probs[small_shape] = count * prob
    return small_probs


def get_counts(size):
    '''
    Count partition shapes of a given sample size.

    Inputs:
        size = sample_size
    Returns:
        dict : shape -> count
    '''
    assert 0 <= size
    cache_file = '{}/counts.{}.json.bz2'.format(TEMP, size)
    if cache_file not in CACHE:
        if os.path.exists(cache_file):
            flat = json_stream_load(cache_file)
            large = {tuple(key): val for key, val in flat}
        else:
            if size == 0:
                large = {(): 1.0}
            else:
                small = get_counts(size - 1)
                large = get_larger_counts(small)
            print 'caching', cache_file
            json_stream_dump(large.iteritems(), cache_file)
        CACHE[cache_file] = large
    return CACHE[cache_file]


def enum_counts(max_size):
    return [get_counts(size) for size in range(1 + max_size)]


def get_log_z(shape):
    return sum(n * math.log(n) for n in shape)


def get_log_Z(counts):
    return numpy.logaddexp.reduce([
        get_log_z(shape) + math.log(count)
        for shape, count in counts.iteritems()
    ])


def get_probs(size):
    counts = get_counts(size).copy()
    for shape, count in counts.iteritems():
        counts[shape] = get_log_z(shape) + math.log(count)
    log_Z = numpy.logaddexp.reduce(counts.values())
    for shape, log_z in counts.iteritems():
        counts[shape] = math.exp(log_z - log_Z)
    return counts


def get_subprobs(size, max_size):
    '''
    Compute probabilities of shapes of partial assignment vectors.

    Inputs:
        size = sample_size
        max_size = dataset_size
    Returns:
        dict : shape -> prob
    '''
    assert 0 <= size
    assert size <= max_size
    cache_file = '{}/subprobs.{}.{}.json.bz2'.format(TEMP, size, max_size)
    if cache_file not in CACHE:
        if os.path.exists(cache_file):
            flat = json_stream_load(cache_file)
            small_probs = {tuple(key): val for key, val in flat}
        else:
            if size == max_size:
                small_probs = get_probs(size)
            else:
                small_counts = get_counts(size)
                large_counts = get_counts(size + 1)
                large_probs = get_subprobs(size + 1, max_size)
                small_probs = get_smaller_probs(
                    small_counts,
                    large_counts,
                    large_probs)
            print 'caching', cache_file
            json_stream_dump(small_probs.iteritems(), cache_file)
        CACHE[cache_file] = small_probs
    return CACHE[cache_file]


def enum_probs(max_size):
    return [get_probs(size) for size in range(max_size + 1)]


@parsable.command
def priors(N=100):
    '''
    Plots different partition priors.
    '''
    X = numpy.array(range(1, N + 1))

    def plot(Y, *args, **kwargs):
        Y = numpy.array(Y)
        Y -= numpy.logaddexp.reduce(Y)
        pyplot.plot(X, Y, *args, **kwargs)

    def crp(alpha):
        assert 0 < alpha
        prob = numpy.zeros(len(X))
        prob[1:] = log(X[1:] - 1)
        prob[0] = log(alpha)
        return prob

    def entropy():
        prob = numpy.zeros(len(X))
        n_log_n = lambda n: n * log(n)
        prob[1:] = n_log_n(X[1:]) - n_log_n(X[1:] - 1)
        return prob

    def plot_crp(alpha):
        plot(crp(eval(alpha)), label='CRP({})'.format(alpha))

    def plot_entropy():
        plot(entropy(), 'k--', linewidth=2, label='low-entropy')

    pyplot.figure(figsize=(8, 4))

    plot_entropy()
    plot_crp('0.01')
    plot_crp('0.1')
    plot_crp('exp(-1)')
    plot_crp('1.0')
    plot_crp('10.0')

    pyplot.title('Posterior Predictive Curves of Clustering Priors')
    pyplot.xlabel('category size')
    pyplot.ylabel('log(probability)')
    pyplot.xscale('log')
    pyplot.legend(loc='best')

    savefig('priors')


def get_pairwise(counts):
    size = sum(iter(counts).next())
    paired = 0.0
    for shape, prob in counts.iteritems():
        paired += prob * sum(n * (n - 1) for n in shape) / (size * (size - 1))
    return paired


@parsable.command
def pairwise(max_size=DEFAULT_MAX_SIZE):
    '''
    Plot probability that two points lie in same cluster,
    as function of data set size.
    '''
    all_counts = enum_probs(max_size)
    sizes = range(2, len(all_counts))
    probs = [get_pairwise(all_counts[i]) for i in sizes]

    pyplot.figure()
    pyplot.plot(sizes, probs, marker='.')
    pyplot.title('\n'.join([
        'Cohabitation probability depends on dataset size',
        '(unlike the CRP or PYP)'
    ]))
    pyplot.xlabel('# objects')
    pyplot.ylabel('P[two random objects in same cluster]')
    pyplot.xscale('log')
    # pyplot.yscale('log')
    pyplot.ylim(0, 1)
    savefig('pairwise')


def get_color_range(size):
    scale = 1.0 / (size - 1.0)
    return [
        (scale * t, 0.5, scale * (size - t - 1))
        for t in range(size)
    ]


def approximate_postpred_correction(subsample_size, dataset_size):
    t = numpy.log(1.0 * dataset_size / subsample_size)
    return t * (0.45 - 0.1 / subsample_size - 0.1 / dataset_size)


def ad_hoc_size_factor(subsample_size, dataset_size):
    return numpy.exp(
        approximate_postpred_correction(subsample_size, dataset_size))


@parsable.command
def postpred(subsample_size=10):
    '''
    Plot posterior predictive probability and approximations,
    fixing subsample size and varying cluster size and dataset size.
    '''
    size = subsample_size
    max_sizes = [size] + [2, 3, 5, 8, 10, 15, 20, 30, 40, 50]
    max_sizes = sorted(set(s for s in max_sizes if s >= size))
    colors = get_color_range(len(max_sizes))
    pyplot.figure(figsize=(12, 8))
    Y_max = 0

    large_counts = get_counts(size)
    for max_size, color in zip(max_sizes, colors):
        large_probs = get_subprobs(size, max_size)
        small_probs = get_subprobs(size - 1, max_size)

        def plot(X, Y, **kwargs):
            pyplot.scatter(
                X, Y,
                color=color,
                edgecolors='none',
                **kwargs)

        plot([], [], label='max_size = {}'.format(max_size))

        max_small_prob = max(small_probs.itervalues())
        for small_shape, small_prob in small_probs.iteritems():
            X = []
            Y = []

            # create a new partition
            n = 1
            large_shape = (1,) + small_shape
            prob = large_probs[large_shape] / large_counts[large_shape]
            X.append(n)
            Y.append(prob)
            singleton_prob = prob

            # add to each existing partition
            for i in range(len(small_shape)):
                n = small_shape[i] + 1
                large_shape = list(small_shape)
                large_shape[i] += 1
                large_shape.sort()
                large_shape = tuple(large_shape)
                prob = large_probs[large_shape] / large_counts[large_shape]
                X.append(n)
                Y.append(prob)

            X = numpy.array(X)
            Y = numpy.array(Y)
            Y /= singleton_prob
            alpha = small_prob / max_small_prob
            plot(X, Y, alpha=alpha)
            Y_max = max(Y_max, max(Y))

    X = numpy.array(range(1, size + 1))

    # entropy
    entropy = numpy.array([
        x * (x / (x - 1.0)) ** (x - 1.0) if x > 1 else 1
        for x in X
    ])
    Y = entropy / entropy.min()
    pyplot.plot(X, Y, 'k--', label='entropy', linewidth=2)

    # CRP
    alpha = math.exp(-1)
    Y = numpy.array([x - 1 if x > 1 else alpha for x in X])
    Y /= Y.min()
    pyplot.plot(X, Y, 'g-', label='CRP(exp(-1))'.format(alpha))

    # ad hoc
    factors = ad_hoc_size_factor(size, numpy.array(max_sizes))
    for factor in factors:
        Y = entropy.copy()
        Y[0] *= factor
        Y /= Y.min()
        pyplot.plot(X, Y, 'r--')
    pyplot.plot([], [], 'r--', label='ad hoc')

    pyplot.yscale('log')
    pyplot.xscale('log')
    pyplot.title(
        'Adding 1 point to subsample of {} points out of {} total'.format(
            size, max_size))
    pyplot.xlabel('cluster size')
    pyplot.ylabel('posterior predictive probability')
    pyplot.xlim(1, size * 1.01)
    pyplot.ylim(1, Y_max * 1.01)
    pyplot.legend(
        prop=font_manager.FontProperties(size=10),
        loc='upper left')
    savefig('postpred')


def true_postpred_correction(subsample_size, dataset_size):
    '''
    Compute true postpred constant according to size-based approximation.
    '''
    large_counts = get_counts(subsample_size)
    large_probs = get_subprobs(subsample_size, dataset_size)
    small_probs = get_subprobs(subsample_size - 1, dataset_size)

    numer = 0
    denom = 0

    for small_shape, small_prob in small_probs.iteritems():
        probs = []

        # create a new partition
        n = 1
        large_shape = (1,) + small_shape
        prob = large_probs[large_shape] / large_counts[large_shape]
        probs.append((n, prob))

        # add to each existing partition
        for i in range(len(small_shape)):
            n = small_shape[i] + 1
            large_shape = list(small_shape)
            large_shape[i] += 1
            large_shape.sort()
            large_shape = tuple(large_shape)
            prob = large_probs[large_shape] / large_counts[large_shape]
            probs.append((n, prob))

        total = sum(prob for _, prob in probs)
        singleton_prob = probs[0][1]
        for n, prob in probs[1:]:
            weight = small_prob * prob / total
            baseline = -math.log(n * (n / (n - 1.0)) ** (n - 1.0))
            correction = math.log(singleton_prob / prob) - baseline
            numer += weight * correction
            denom += weight

    return numer / denom if denom > 0 else 1.0


@parsable.command
def dataprob(subsample_size=10, dataset_size=50):
    '''
    Plot data prob approximation.

    This tests the accuracy of LowEntropy.score_counts(...).
    '''
    true_probs = get_subprobs(subsample_size, dataset_size)
    naive_probs = get_probs(subsample_size)
    shapes = true_probs.keys()

    # apply ad hoc size factor
    approx_probs = naive_probs.copy()
    factor = ad_hoc_size_factor(subsample_size, dataset_size)
    print 'factor =', factor
    for shape in shapes:
        approx_probs[shape] *= factor ** (len(shape) - 2)

    X = numpy.array([true_probs[shape] for shape in shapes])
    Y0 = numpy.array([naive_probs[shape] for shape in shapes])
    Y1 = numpy.array([approx_probs[shape] for shape in shapes])

    pyplot.figure()
    pyplot.scatter(X, Y0, color='blue', edgecolors='none', label='naive')
    pyplot.scatter(X, Y1, color='red', edgecolors='none', label='approx')
    pyplot.xlabel('true probability')
    pyplot.ylabel('approximation')
    LB = min(X.min(), Y0.min(), Y1.min())
    UB = max(X.max(), Y0.max(), Y1.max())
    pyplot.xlim(LB, UB)
    pyplot.ylim(LB, UB)
    pyplot.plot([LB, UB], [LB, UB], 'k--')
    pyplot.xscale('log')
    pyplot.yscale('log')
    pyplot.title('\n'.join([
        'Approximate data probability',
        'subsample_size = {}, dataset_size = {}'.format(
            subsample_size,
            dataset_size),
    ]))
    pyplot.legend(
        prop=font_manager.FontProperties(size=10),
        loc='lower right')
    savefig('dataprob')


def true_dataprob_correction(subsample_size, dataset_size):
    '''
    Compute true normalization correction.
    '''
    naive_probs = get_probs(subsample_size)
    factor = ad_hoc_size_factor(subsample_size, dataset_size)
    Z = sum(
        prob * factor ** (len(shape) - 1)
        for shape, prob in naive_probs.iteritems()
    )
    return -math.log(Z)


def approximate_dataprob_correction(subsample_size, dataset_size):
    n = math.log(subsample_size)
    N = math.log(dataset_size)
    return 0.061 * n * (n - N) * (n + N) ** 0.75


@parsable.command
def normalization(max_size=DEFAULT_MAX_SIZE):
    '''
    Plot approximation to partition function of low-entropy clustering
    distribution for various set sizes.
    '''
    pyplot.figure()

    all_counts = enum_counts(max_size)
    sizes = numpy.array(range(1, 1 + max_size))
    log_Z = numpy.array([
        get_log_Z(all_counts[size]) for size in sizes
    ])
    log_z_max = numpy.array([get_log_z([size]) for size in sizes])

    coeffs = [
        (log_Z[i] / log_z_max[i] - 1.0) * sizes[i] ** 0.75
        for i in [1, -1]
    ]
    coeffs += [0.27, 0.275, 0.28]
    for coeff in coeffs:
        print coeff
        approx = log_z_max * (1 + coeff * sizes ** -0.75)
        X = sizes ** -0.75
        Y = (log_Z - approx) / log_Z
        pyplot.plot(X, Y, marker='.', label='coeff = {}'.format(coeff))
    pyplot.xlim(0, 1)
    pyplot.xlabel('1 / size')
    pyplot.ylabel('approx error')
    pyplot.title(
        'log(Z) ~ log(z_max) * (1 + coeff * size ** -0.75)')
    pyplot.legend(loc='best')
    savefig('normalization')


@parsable.command
def approximations(max_size=DEFAULT_MAX_SIZE):
    '''
    Plot both main approximations for many (subsample, dataset) sizes:
    (1) normalization constant, and
    (2) postpred size factor
    '''
    sizes = [1, 2, 3, 4, 5, 7, 10, 15, 20, 30, 40, 50, 60]
    sizes = [size for size in sizes if size <= max_size]
    keys = [(x, y) for x in sizes for y in sizes if x <= y]

    truth1 = {}
    truth2 = {}
    approx1 = {}
    approx2 = {}
    for key in keys:
        size, max_size = key

        # postpred correction
        if size > 1:
            truth1[key] = true_postpred_correction(size, max_size)
            approx1[key] = approximate_postpred_correction(size, max_size)

        # normalization correction
        truth2[key] = true_dataprob_correction(size, max_size)
        approx2[key] = approximate_dataprob_correction(size, max_size)

    fig, (ax1, ax2) = pyplot.subplots(2, 1, sharex=True, figsize=(12, 8))
    ax1.set_title('Approximation accuracies of postpred and dataprob')
    ax2.set_ylabel('log(Z correction)')
    ax1.set_ylabel('log(singleton postpred correction)')
    ax2.set_xlabel('subsample size')
    ax1.set_xlim(min(sizes) * 0.95, max(sizes) * 1.05)
    ax1.set_xscale('log')
    ax2.set_xscale('log')

    def plot(ax, X, y, values, *args, **kwargs):
        Y = [values[x, y] for x in X if (x, y) in values]
        X = [x for x in X if (x, y) in values]
        ax.plot(X, Y, *args, alpha=0.5, marker='.', **kwargs)

    for max_size in sizes:
        X = [n for n in sizes if n <= max_size]
        plot(ax1, X, max_size, truth1, 'k-')
        plot(ax1, X, max_size, approx1, 'r-')
        plot(ax2, X, max_size, truth2, 'k-')
        plot(ax2, X, max_size, approx2, 'r-')

    plot(ax1, [], None, {}, 'r-', label='approximation')
    plot(ax1, [], None, {}, 'k-', label='truth')
    ax1.legend(loc='upper right')

    savefig('approximations')


@parsable.command
def fastlog():
    '''
    Plot accuracy of fastlog term in cluster_add_score.
    '''
    X = numpy.array([2.0 ** i for i in range(20 + 1)])
    Y0 = numpy.array([x * math.log(1. + 1. / x) for x in X])
    Y1 = numpy.array([x * fast_log(1. + 1. / x) for x in X])
    Y2 = numpy.array([1.0 for x in X])

    fig, (ax1, ax2) = pyplot.subplots(2, 1, sharex=True)

    ax1.plot(X, Y0, 'ko', label='math.log')
    ax1.plot(X, Y1, 'r-', label='lp.special.fast_log')
    ax1.plot(X, Y2, 'b-', label='asymptote')
    ax2.plot(X, numpy.abs(Y1 - Y0), 'r-', label='lp.special.fast_log')
    ax2.plot(X, numpy.abs(Y2 - Y0), 'b-', label='asymptote')

    ax1.set_title('lp.special.fast_log approximation')
    ax1.set_ylabel('n log(1 + 1 / n)')
    ax2.set_ylabel('approximation error')
    ax2.set_xlabel('n')
    ax1.set_xscale('log')
    ax2.set_xscale('log')
    ax2.set_yscale('log')
    ax1.legend(loc='best')
    ax2.legend(loc='best')
    savefig('fastlog')


def number_table(numbers, width=5):
    lines = []
    line = ''
    for i, number in enumerate(numbers):
        if i % width == 0:
            if line:
                lines.append(line + ',')
            line = '    %0.8f' % number
        else:
            line += ', %0.8f' % number
    if line:
        lines.append(line)
    return '\n'.join(lines)


@parsable.command
def code(max_size=DEFAULT_MAX_SIZE):
    '''
    Generate C++ code for clustering partition function.
    '''
    all_counts = enum_counts(max_size)
    sizes = range(1 + max_size)
    log_Z = [
        get_log_Z(all_counts[size]) if size else 0
        for size in sizes
    ]
    size = sizes[-1]
    coeff = (log_Z[-1] / get_log_z([size]) - 1.0) * size ** 0.75

    print '# Insert this in src/clustering.cc:'
    lines = [
        '// this code was generated by derivations/clustering.py',
        'static const float log_partition_function_table[%d] =' %
        (max_size + 1),
        '{',
        number_table(log_Z),
        '};',
        '',
        '// this code was generated by derivations/clustering.py',
        'template<class count_t>',
        'float Clustering<count_t>::LowEntropy::log_partition_function (',
        '        count_t sample_size) const',
        '{',
        '    // TODO incorporate dataset_size for higher accuracy',
        '    count_t n = sample_size;',
        '    if (n < %d) {' % (max_size + 1),
        '        return log_partition_function_table[n];',
        '    } else {',
        '        float coeff = %0.8ff;' % coeff,
        '        float log_z_max = n * fast_log(n);',
        '        return log_z_max * (1.f + coeff * powf(n, -0.75f));',
        '    }',
        '}',
    ]
    print '\n'.join(lines)
    print

    print '# Insert this in distributions/dbg/clustering.py:'
    lines = [
        '# this code was generated by derivations/clustering.py',
        'log_partition_function_table = [',
        number_table(log_Z),
        ']',
        '',
        '',
        '# this code was generated by derivations/clustering.py',
        'def log_partition_function(sample_size):',
        '    # TODO incorporate dataset_size for higher accuracy',
        '    n = sample_size',
        '    if n < %d:' % (max_size + 1),
        '        return LowEntropy.log_partition_function_table[n]',
        '    else:',
        '        coeff = %0.8f' % coeff,
        '        log_z_max = n * log(n)',
        '        return log_z_max * (1.0 + coeff * n ** -0.75)',
    ]
    print '\n'.join(lines)
    print


@parsable.command
def plots():
    '''
    Generate all plots.
    '''
    priors()
    pairwise()
    postpred()
    dataprob()
    approximations()
    normalization()
    fastlog()


if __name__ == '__main__':
    parsable.dispatch()



# Copyright (c) 2014, Salesforce.com, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# - Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# - Neither the name of Salesforce.com nor the names of its contributors
#   may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import math
import numpy
import numpy.random
import scipy.stats
import functools
from collections import defaultdict
from nose import SkipTest
from nose.tools import (
    assert_true,
    assert_in,
    assert_is_instance,
    assert_not_equal,
    assert_greater,
)
from distributions.dbg.random import sample_discrete
from distributions.util import (
    scores_to_probs,
    density_goodness_of_fit,
    discrete_goodness_of_fit,
)
from distributions.tests.util import (
    assert_hasattr,
    assert_close,
    assert_all_close,
    list_models,
    import_model,
    seed_all,
)

try:
    import distributions.io.schema_pb2
    has_protobuf = True
except ImportError:
    has_protobuf = False

DATA_COUNT = 20
SAMPLE_COUNT = 1000
MIN_GOODNESS_OF_FIT = 1e-3

MODULES = {
    '{flavor}.models.{name}'.format(**spec): import_model(spec)
    for spec in list_models()
}

IS_FAST = {'dbg': False, 'hp': True, 'lp': True}


def model_is_fast(model):
    flavor = model.__name__.split('.')[1]
    return IS_FAST[flavor]


def iter_examples(module):
    assert_hasattr(module, 'EXAMPLES')
    EXAMPLES = module.EXAMPLES
    assert_is_instance(EXAMPLES, list)
    assert_true(EXAMPLES, 'no examples provided')
    for i, EXAMPLE in enumerate(EXAMPLES):
        print 'example {}/{}'.format(1 + i, len(EXAMPLES))
        assert_in('shared', EXAMPLE)
        assert_in('values', EXAMPLE)
        values = EXAMPLE['values']
        assert_is_instance(values, list)
        count = len(values)
        assert_true(
            count >= 7,
            'Add more example values (expected >= 7, found {})'.format(count))
        yield EXAMPLE


def for_each_model(*filters):
    '''
    Run one test per Model, filtering out inappropriate Models for test.
    '''
    def filtered(test_fun):

        @functools.wraps(test_fun)
        def test_one_model(name):
            module = MODULES[name]
            assert_hasattr(module, 'Shared')
            for EXAMPLE in iter_examples(module):
                test_fun(module, EXAMPLE)

        @functools.wraps(test_fun)
        def test_all_models():
            for name in MODULES:
                module = MODULES[name]
                if all(f(module) for f in filters):
                    yield test_one_model, name

        return test_all_models
    return filtered


@for_each_model()
def test_value(module, EXAMPLE):
    assert_hasattr(module, 'Value')
    assert_is_instance(module.Value, type)

    values = EXAMPLE['values']
    for value in values:
        assert_is_instance(value, module.Value)


@for_each_model()
def test_shared(module, EXAMPLE):
    assert_hasattr(module, 'Shared')
    assert_is_instance(module.Shared, type)

    shared1 = module.Shared.from_dict(EXAMPLE['shared'])
    shared2 = module.Shared.from_dict(EXAMPLE['shared'])
    assert_close(shared1.dump(), EXAMPLE['shared'])

    values = EXAMPLE['values']
    seed_all(0)
    for value in values:
        shared1.add_value(value)
    seed_all(0)
    for value in values:
        shared2.add_value(value)
    assert_close(shared1.dump(), shared2.dump())

    for value in values:
        shared1.remove_value(value)
    assert_close(shared1.dump(), EXAMPLE['shared'])


@for_each_model()
def test_group(module, EXAMPLE):
    assert_hasattr(module, 'Group')
    assert_is_instance(module.Group, type)

    shared = module.Shared.from_dict(EXAMPLE['shared'])
    values = EXAMPLE['values']
    for value in values:
        shared.add_value(value)

    group1 = module.Group()
    group1.init(shared)
    for value in values:
        group1.add_value(shared, value)
    group2 = module.Group.from_values(shared, values)
    assert_close(group1.dump(), group2.dump())

    group = module.Group.from_values(shared, values)
    dumped = group.dump()
    group.init(shared)
    group.load(dumped)
    assert_close(group.dump(), dumped)

    for value in values:
        group2.remove_value(shared, value)
    assert_not_equal(group1, group2)
    group2.merge(shared, group1)

    for value in values:
        group1.score_value(shared, value)
    for _ in xrange(10):
        value = group1.sample_value(shared)
        group1.score_value(shared, value)
        module.sample_group(shared, 10)
    group1.score_data(shared)
    group2.score_data(shared)


@for_each_model(lambda module: hasattr(module.Shared, 'load_protobuf'))
def test_protobuf(module, EXAMPLE):
    if not has_protobuf:
        raise SkipTest('protobuf not available')
    shared = module.Shared.from_dict(EXAMPLE['shared'])
    values = EXAMPLE['values']
    Message = getattr(distributions.io.schema_pb2, module.NAME)

    message = Message.Shared()
    shared.dump_protobuf(message)
    shared2 = module.Shared()
    shared2.load_protobuf(message)
    assert_close(shared2.dump(), shared.dump())

    message.Clear()
    dumped = shared.dump()
    module.Shared.to_protobuf(dumped, message)
    assert_close(module.Shared.from_protobuf(message), dumped)

    if hasattr(module.Group, 'load_protobuf'):
        for value in values:
            shared.add_value(value)
        group = module.Group.from_values(shared, values)

        message = Message.Group()
        group.dump_protobuf(message)
        group2 = module.Group()
        group2.load_protobuf(message)
        assert_close(group2.dump(), group.dump())

        message.Clear()
        dumped = group.dump()
        module.Group.to_protobuf(dumped, message)
        assert_close(module.Group.from_protobuf(message), dumped)


@for_each_model()
def test_add_remove(module, EXAMPLE):
    # Test group_add_value, group_remove_value, score_data, score_value

    shared = module.Shared.from_dict(EXAMPLE['shared'])
    shared.realize()

    values = []
    group = module.Group.from_values(shared)
    score = 0.0
    assert_close(group.score_data(shared), score, err_msg='p(empty) != 1')

    for _ in range(DATA_COUNT):
        value = group.sample_value(shared)
        values.append(value)
        score += group.score_value(shared, value)
        group.add_value(shared, value)

    group_all = module.Group.from_dict(group.dump())
    assert_close(
        score,
        group.score_data(shared),
        err_msg='p(x1,...,xn) != p(x1) p(x2|x1) p(xn|...)')

    numpy.random.shuffle(values)

    for value in values:
        group.remove_value(shared, value)

    group_empty = module.Group.from_values(shared)
    assert_close(
        group.dump(),
        group_empty.dump(),
        err_msg='group + values - values != group')

    numpy.random.shuffle(values)
    for value in values:
        group.add_value(shared, value)
    assert_close(
        group.dump(),
        group_all.dump(),
        err_msg='group - values + values != group')


@for_each_model()
def test_add_repeated(module, EXAMPLE):
    # Test add_repeated value vs n * add
    shared = module.Shared.from_dict(EXAMPLE['shared'])
    shared.realize()
    for value in EXAMPLE['values']:
        group = module.Group.from_values(shared)
        for _ in range(DATA_COUNT):
            group.add_value(shared, value)

        group_repeated = module.Group.from_values(shared)
        group_repeated.add_repeated_value(shared, value, count=DATA_COUNT)
        assert_close(
            group.dump(),
            group_repeated.dump(),
            err_msg='n * add_value != add_repeated_value n')


@for_each_model()
def test_add_merge(module, EXAMPLE):
    # Test group_add_value, group_merge
    shared = module.Shared.from_dict(EXAMPLE['shared'])
    values = EXAMPLE['values'][:]
    for value in values:
        shared.add_value(value)

    numpy.random.shuffle(values)
    group = module.Group.from_values(shared, values)

    for i in xrange(len(values) + 1):
        numpy.random.shuffle(values)
        group1 = module.Group.from_values(shared, values[:i])
        group2 = module.Group.from_values(shared, values[i:])
        group1.merge(shared, group2)
        assert_close(group.dump(), group1.dump())


@for_each_model()
def test_group_merge(module, EXAMPLE):
    shared = module.Shared.from_dict(EXAMPLE['shared'])
    shared.realize()
    group1 = module.Group.from_values(shared)
    group2 = module.Group.from_values(shared)
    expected = module.Group.from_values(shared)
    actual = module.Group.from_values(shared)
    for _ in xrange(100):
        value = expected.sample_value(shared)
        expected.add_value(shared, value)
        group1.add_value(shared, value)

        value = expected.sample_value(shared)
        expected.add_value(shared, value)
        group2.add_value(shared, value)

        actual.load(group1.dump())
        actual.merge(shared, group2)
        assert_close(actual.dump(), expected.dump())


@for_each_model(lambda module: module.Value in [bool, int])
def test_group_allows_debt(module, EXAMPLE):
    # Test that group.add_value can safely go into data debt
    shared = module.Shared.from_dict(EXAMPLE['shared'])
    shared.realize()
    values = []
    group1 = module.Group.from_values(shared, values)
    for _ in range(DATA_COUNT):
        value = group1.sample_value(shared)
        values.append(value)
        group1.add_value(shared, value)

    group2 = module.Group.from_values(shared)
    pos_values = [(v, +1) for v in values]
    neg_values = [(v, -1) for v in values]
    signed_values = pos_values * 3 + neg_values * 2
    numpy.random.shuffle(signed_values)
    for value, sign in signed_values:
        if sign > 0:
            group2.add_value(shared, value)
        else:
            group2.remove_value(shared, value)

    assert_close(group1.dump(), group2.dump())


@for_each_model()
def test_sample_seed(module, EXAMPLE):
    shared = module.Shared.from_dict(EXAMPLE['shared'])

    seed_all(0)
    group1 = module.Group.from_values(shared)
    values1 = [group1.sample_value(shared) for _ in xrange(DATA_COUNT)]

    seed_all(0)
    group2 = module.Group.from_values(shared)
    values2 = [group2.sample_value(shared) for _ in xrange(DATA_COUNT)]

    assert_close(values1, values2, err_msg='values')


@for_each_model()
def test_sample_value(module, EXAMPLE):
    seed_all(0)
    shared = module.Shared.from_dict(EXAMPLE['shared'])
    shared.realize()
    for values in [[], EXAMPLE['values']]:
        group = module.Group.from_values(shared, values)
        samples = [group.sample_value(shared) for _ in xrange(SAMPLE_COUNT)]
        if module.Value in [bool, int]:
            probs_dict = {
                value: math.exp(group.score_value(shared, value))
                for value in set(samples)
            }
            gof = discrete_goodness_of_fit(samples, probs_dict, plot=True)
        elif module.Value == float:
            probs = numpy.exp([
                group.score_value(shared, value)
                for value in samples
            ])
            gof = density_goodness_of_fit(samples, probs, plot=True)
        else:
            raise SkipTest('Not implemented for {}'.format(module.Value))
        print '{} gof = {:0.3g}'.format(module.__name__, gof)
        assert_greater(gof, MIN_GOODNESS_OF_FIT)


@for_each_model()
def test_sample_group(module, EXAMPLE):
    seed_all(0)
    SIZE = 2
    shared = module.Shared.from_dict(EXAMPLE['shared'])
    shared.realize()
    for values in [[], EXAMPLE['values']]:
        if module.Value in [bool, int]:
            samples = []
            probs_dict = {}
            for _ in xrange(SAMPLE_COUNT):
                values = module.sample_group(shared, SIZE)
                sample = tuple(values)
                samples.append(sample)
                group = module.Group.from_values(shared, values)
                probs_dict[sample] = math.exp(group.score_data(shared))
            gof = discrete_goodness_of_fit(samples, probs_dict, plot=True)
        else:
            raise SkipTest('Not implemented for {}'.format(module.Value))
        print '{} gof = {:0.3g}'.format(module.__name__, gof)
        assert_greater(gof, MIN_GOODNESS_OF_FIT)


def _append_ss(group, aggregator):
    ss = group.dump()
    for key, val in ss.iteritems():
        if isinstance(val, list):
            for i, v in enumerate(val):
                aggregator['{}_{}'.format(key, i)].append(v)
        elif isinstance(val, dict):
            for k, v in val.iteritems():
                aggregator['{}_{}'.format(key, k)].append(v)
        else:
            aggregator[key].append(val)


def sample_marginal_conditional(module, shared, value_count):
    values = module.sample_group(shared, value_count)
    group = module.Group.from_values(shared, values)
    return group


def sample_successive_conditional(module, shared, group, value_count):
    sampler = module.Sampler()
    sampler.init(shared, group)
    values = [sampler.eval(shared) for _ in xrange(value_count)]
    new_group = module.Group.from_values(shared, values)
    return new_group


@for_each_model(model_is_fast)
def test_joint(module, EXAMPLE):
    # \cite{geweke04getting}
    seed_all(0)
    SIZE = 10
    SKIP = 100
    shared = module.Shared.from_dict(EXAMPLE['shared'])
    shared.realize()
    marginal_conditional_samples = defaultdict(lambda: [])
    successive_conditional_samples = defaultdict(lambda: [])
    cond_group = sample_marginal_conditional(module, shared, SIZE)
    for _ in xrange(SAMPLE_COUNT):
        marg_group = sample_marginal_conditional(module, shared, SIZE)
        _append_ss(marg_group, marginal_conditional_samples)

        for __ in range(SKIP):
            cond_group = sample_successive_conditional(
                module,
                shared,
                cond_group,
                SIZE)
        _append_ss(cond_group, successive_conditional_samples)
    for key in marginal_conditional_samples.keys():
        gof = scipy.stats.ttest_ind(
            marginal_conditional_samples[key],
            successive_conditional_samples[key])[1]
        print '{}:{} gof = {:0.3g}'.format(module.__name__, key, gof)
        if not numpy.isfinite(gof):
            raise SkipTest('Test fails with gof = {}'.format(gof))
        assert_greater(gof, MIN_GOODNESS_OF_FIT)


@for_each_model(lambda module: hasattr(module.Shared, 'scorer_create'))
def test_scorer(module, EXAMPLE):
    shared = module.Shared.from_dict(EXAMPLE['shared'])
    values = EXAMPLE['values']

    group = module.Group.from_values(shared)
    scorer1 = shared.scorer_create()
    scorer2 = shared.scorer_create(group)
    for value in values:
        score1 = shared.scorer_eval(scorer1, value)
        score2 = shared.scorer_eval(scorer2, value)
        score3 = group.score_value(shared, value)
        assert_all_close([score1, score2, score3])


@for_each_model(lambda module: hasattr(module, 'Mixture'))
def test_mixture_runs(module, EXAMPLE):
    shared = module.Shared.from_dict(EXAMPLE['shared'])
    values = EXAMPLE['values']

    mixture = module.Mixture()
    for value in values:
        shared.add_value(value)
        mixture.append(module.Group.from_values(shared, [value]))
    mixture.init(shared)

    groupids = []
    for value in values:
        scores = numpy.zeros(len(mixture), dtype=numpy.float32)
        mixture.score_value(shared, value, scores)
        probs = scores_to_probs(scores)
        groupid = sample_discrete(probs)
        mixture.add_value(shared, groupid, value)
        groupids.append(groupid)

    mixture.add_group(shared)
    assert len(mixture) == len(values) + 1
    scores = numpy.zeros(len(mixture), dtype=numpy.float32)

    for value, groupid in zip(values, groupids):
        mixture.remove_value(shared, groupid, value)

    mixture.remove_group(shared, 0)
    mixture.remove_group(shared, len(mixture) - 1)
    assert len(mixture) == len(values) - 1

    for value in values:
        scores = numpy.zeros(len(mixture), dtype=numpy.float32)
        mixture.score_value(shared, value, scores)
        probs = scores_to_probs(scores)
        groupid = sample_discrete(probs)
        mixture.add_value(shared, groupid, value)


@for_each_model(lambda module: hasattr(module, 'Mixture'))
def test_mixture_score(module, EXAMPLE):
    shared = module.Shared.from_dict(EXAMPLE['shared'])
    values = EXAMPLE['values']
    for value in values:
        shared.add_value(value)

    groups = [module.Group.from_values(shared, [value]) for value in values]
    mixture = module.Mixture()
    for group in groups:
        mixture.append(group)
    mixture.init(shared)

    def check_score_value(value):
        expected = [group.score_value(shared, value) for group in groups]
        actual = numpy.zeros(len(mixture), dtype=numpy.float32)
        noise = numpy.random.randn(len(actual))
        actual += noise
        mixture.score_value(shared, value, actual)
        actual -= noise
        assert_close(actual, expected, err_msg='score_value {}'.format(value))
        another = [
            mixture.score_value_group(shared, i, value)
            for i in xrange(len(groups))
        ]
        assert_close(
            another,
            expected,
            err_msg='score_value_group {}'.format(value))
        return actual

    def check_score_data():
        expected = sum(group.score_data(shared) for group in groups)
        actual = mixture.score_data(shared)
        assert_close(actual, expected, err_msg='score_data')

    print 'init'
    for value in values:
        check_score_value(value)
    check_score_data()

    print 'adding'
    groupids = []
    for value in values:
        scores = check_score_value(value)
        probs = scores_to_probs(scores)
        groupid = sample_discrete(probs)
        groups[groupid].add_value(shared, value)
        mixture.add_value(shared, groupid, value)
        groupids.append(groupid)
        check_score_data()

    print 'removing'
    for value, groupid in zip(values, groupids):
        groups[groupid].remove_value(shared, value)
        mixture.remove_value(shared, groupid, value)
        scores = check_score_value(value)
        check_score_data()



# Copyright (c) 2014, Salesforce.com, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# - Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# - Neither the name of Salesforce.com nor the names of its contributors
#   may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

cimport _dd
import _dd

from distributions.mixins import SharedMixin, GroupIoMixin, SharedIoMixin


NAME = 'DirichletDiscrete'
EXAMPLES = [
    {
        'shared': {'alphas': [0.5, 0.5, 0.5, 0.5]},
        'values': [0, 1, 0, 2, 0, 1, 0],
    },
    {
        'shared': {'alphas': [1.0, 4.0]},
        'values': [0, 1, 1, 1, 1, 0, 1],
    },
    {
        'shared': {'alphas': [2.0 / n for n in xrange(1, 21)]},
        'values': range(20),
    },
]
Value = int


cdef class _Shared(_dd.Shared):
    def load(self, raw):
        alphas = raw['alphas']
        cdef int dim = len(alphas)
        self.ptr.dim = dim
        cdef int i
        for i in xrange(dim):
            self.ptr.alphas[i] = float(alphas[i])

    def dump(self):
        alphas = []
        cdef int i
        for i in xrange(self.ptr.dim):
            alphas.append(float(self.ptr.alphas[i]))
        return {'alphas': alphas}

    def load_protobuf(self, message):
        cdef int dim = len(message.alphas)
        self.ptr.dim = dim
        cdef int i
        for i in xrange(self.ptr.dim):
            self.ptr.alphas[i] = message.alphas[i]

    def dump_protobuf(self, message):
        message.Clear()
        cdef int i
        for i in xrange(self.ptr.dim):
            message.alphas.append(float(self.ptr.alphas[i]))


class Shared(_Shared, SharedMixin, SharedIoMixin):
    pass


cdef class _Group(_dd.Group):
    cdef int dim  # only required for dumping

    def __cinit__(self):
        self.dim = 0

    def load(self, dict raw):
        counts = raw['counts']
        self.dim = len(counts)
        self.ptr.count_sum = 0
        cdef int i
        for i in xrange(self.dim):
            self.ptr.count_sum += counts[i]
            self.ptr.counts[i] = counts[i]

    def dump(self):
        counts = []
        cdef int i
        for i in xrange(self.dim):
            counts.append(self.ptr.counts[i])
        return {'counts': counts}

    def init(self, _dd.Shared shared):
        self.dim = shared.ptr.dim
        _dd.Group.init(self, shared)


class Group(_Group, GroupIoMixin):
    pass


class Sampler(_dd.Sampler):
    pass


Mixture = _dd.Mixture
sample_group = _dd.sample_group



# Generated by the protocol buffer compiler.  DO NOT EDIT!

from google.protobuf import descriptor
from google.protobuf import message
from google.protobuf import reflection
# @@protoc_insertion_point(imports)



DESCRIPTOR = descriptor.FileDescriptor(
  name='distributions/io/schema.proto',
  package='protobuf.distributions',
  serialized_pb='\n\x1d\x64istributions/io/schema.proto\x12\x16protobuf.distributions\"\xdd\x01\n\nClustering\x12@\n\npitman_yor\x18\x01 \x01(\x0b\x32,.protobuf.distributions.Clustering.PitmanYor\x12\x42\n\x0blow_entropy\x18\x02 \x01(\x0b\x32-.protobuf.distributions.Clustering.LowEntropy\x1a%\n\tPitmanYor\x12\r\n\x05\x61lpha\x18\x01 \x02(\x02\x12\t\n\x01\x64\x18\x02 \x02(\x02\x1a\"\n\nLowEntropy\x12\x14\n\x0c\x64\x61taset_size\x18\x01 \x02(\x04\"\x87\x01\n\rBetaBernoulli\x1a%\n\x06Shared\x12\r\n\x05\x61lpha\x18\x01 \x02(\x02\x12\x0c\n\x04\x62\x65ta\x18\x02 \x02(\x02\x1a(\n\tGridPrior\x12\r\n\x05\x61lpha\x18\x01 \x03(\x02\x12\x0c\n\x04\x62\x65ta\x18\x02 \x03(\x02\x1a%\n\x05Group\x12\r\n\x05heads\x18\x01 \x02(\x04\x12\r\n\x05tails\x18\x02 \x02(\x04\"b\n\x11\x44irichletDiscrete\x1a\x18\n\x06Shared\x12\x0e\n\x06\x61lphas\x18\x01 \x03(\x02\x1a\x1a\n\tGridPrior\x12\r\n\x05\x61lpha\x18\x01 \x03(\x02\x1a\x17\n\x05Group\x12\x0e\n\x06\x63ounts\x18\x01 \x03(\x04\"\xc3\x01\n\x18\x44irichletProcessDiscrete\x1aU\n\x06Shared\x12\r\n\x05gamma\x18\x01 \x02(\x02\x12\r\n\x05\x61lpha\x18\x02 \x02(\x02\x12\x0e\n\x06values\x18\x03 \x03(\r\x12\r\n\x05\x62\x65tas\x18\x04 \x03(\x02\x12\x0e\n\x06\x63ounts\x18\x05 \x03(\x04\x1a)\n\tGridPrior\x12\r\n\x05gamma\x18\x01 \x03(\x02\x12\r\n\x05\x61lpha\x18\x02 \x03(\x02\x1a%\n\x05Group\x12\x0c\n\x04keys\x18\x01 \x03(\r\x12\x0e\n\x06values\x18\x02 \x03(\x04\"\xc1\x01\n\x18PitmanYorProcessDiscrete\x1a\x32\n\x06Shared\x12\r\n\x05\x61lpha\x18\x01 \x02(\x02\x12\t\n\x01\x64\x18\x02 \x03(\x02\x12\x0e\n\x06\x63ounts\x18\x03 \x03(\x04\x1aJ\n\tGridPrior\x12=\n\x07\x61lpha_d\x18\x01 \x03(\x0b\x32,.protobuf.distributions.Clustering.PitmanYor\x1a%\n\x05Group\x12\x0c\n\x04keys\x18\x01 \x03(\r\x12\x0e\n\x06values\x18\x02 \x03(\x04\"\x9e\x01\n\x0cGammaPoisson\x1a)\n\x06Shared\x12\r\n\x05\x61lpha\x18\x01 \x02(\x02\x12\x10\n\x08inv_beta\x18\x02 \x02(\x02\x1a,\n\tGridPrior\x12\r\n\x05\x61lpha\x18\x01 \x03(\x02\x12\x10\n\x08inv_beta\x18\x02 \x03(\x02\x1a\x35\n\x05Group\x12\r\n\x05\x63ount\x18\x01 \x02(\x04\x12\x0b\n\x03sum\x18\x02 \x02(\x04\x12\x10\n\x08log_prod\x18\x03 \x02(\x02\"\xa2\x01\n\x14\x42\x65taNegativeBinomial\x1a\x30\n\x06Shared\x12\r\n\x05\x61lpha\x18\x01 \x02(\x02\x12\x0c\n\x04\x62\x65ta\x18\x02 \x02(\x02\x12\t\n\x01r\x18\x03 \x02(\x04\x1a\x33\n\tGridPrior\x12\r\n\x05\x61lpha\x18\x01 \x03(\x02\x12\x0c\n\x04\x62\x65ta\x18\x02 \x03(\x02\x12\t\n\x01r\x18\x03 \x03(\x04\x1a#\n\x05Group\x12\r\n\x05\x63ount\x18\x01 \x02(\x04\x12\x0b\n\x03sum\x18\x02 \x02(\x04\"\xdf\x01\n\x12NormalInverseChiSq\x1a@\n\x06Shared\x12\n\n\x02mu\x18\x01 \x02(\x02\x12\r\n\x05kappa\x18\x02 \x02(\x02\x12\x0f\n\x07sigmasq\x18\x03 \x02(\x02\x12\n\n\x02nu\x18\x04 \x02(\x02\x1a\x43\n\tGridPrior\x12\n\n\x02mu\x18\x01 \x03(\x02\x12\r\n\x05kappa\x18\x02 \x03(\x02\x12\x0f\n\x07sigmasq\x18\x03 \x03(\x02\x12\n\n\x02nu\x18\x04 \x03(\x02\x1a\x42\n\x05Group\x12\r\n\x05\x63ount\x18\x01 \x02(\x04\x12\x0c\n\x04mean\x18\x02 \x02(\x02\x12\x1c\n\x14\x63ount_times_variance\x18\x03 \x02(\x02')




_CLUSTERING_PITMANYOR = descriptor.Descriptor(
  name='PitmanYor',
  full_name='protobuf.distributions.Clustering.PitmanYor',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='alpha', full_name='protobuf.distributions.Clustering.PitmanYor.alpha', index=0,
      number=1, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='d', full_name='protobuf.distributions.Clustering.PitmanYor.d', index=1,
      number=2, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=206,
  serialized_end=243,
)

_CLUSTERING_LOWENTROPY = descriptor.Descriptor(
  name='LowEntropy',
  full_name='protobuf.distributions.Clustering.LowEntropy',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='dataset_size', full_name='protobuf.distributions.Clustering.LowEntropy.dataset_size', index=0,
      number=1, type=4, cpp_type=4, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=245,
  serialized_end=279,
)

_CLUSTERING = descriptor.Descriptor(
  name='Clustering',
  full_name='protobuf.distributions.Clustering',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='pitman_yor', full_name='protobuf.distributions.Clustering.pitman_yor', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='low_entropy', full_name='protobuf.distributions.Clustering.low_entropy', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_CLUSTERING_PITMANYOR, _CLUSTERING_LOWENTROPY, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=58,
  serialized_end=279,
)


_BETABERNOULLI_SHARED = descriptor.Descriptor(
  name='Shared',
  full_name='protobuf.distributions.BetaBernoulli.Shared',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='alpha', full_name='protobuf.distributions.BetaBernoulli.Shared.alpha', index=0,
      number=1, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='beta', full_name='protobuf.distributions.BetaBernoulli.Shared.beta', index=1,
      number=2, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=299,
  serialized_end=336,
)

_BETABERNOULLI_GRIDPRIOR = descriptor.Descriptor(
  name='GridPrior',
  full_name='protobuf.distributions.BetaBernoulli.GridPrior',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='alpha', full_name='protobuf.distributions.BetaBernoulli.GridPrior.alpha', index=0,
      number=1, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='beta', full_name='protobuf.distributions.BetaBernoulli.GridPrior.beta', index=1,
      number=2, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=338,
  serialized_end=378,
)

_BETABERNOULLI_GROUP = descriptor.Descriptor(
  name='Group',
  full_name='protobuf.distributions.BetaBernoulli.Group',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='heads', full_name='protobuf.distributions.BetaBernoulli.Group.heads', index=0,
      number=1, type=4, cpp_type=4, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='tails', full_name='protobuf.distributions.BetaBernoulli.Group.tails', index=1,
      number=2, type=4, cpp_type=4, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=380,
  serialized_end=417,
)

_BETABERNOULLI = descriptor.Descriptor(
  name='BetaBernoulli',
  full_name='protobuf.distributions.BetaBernoulli',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[_BETABERNOULLI_SHARED, _BETABERNOULLI_GRIDPRIOR, _BETABERNOULLI_GROUP, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=282,
  serialized_end=417,
)


_DIRICHLETDISCRETE_SHARED = descriptor.Descriptor(
  name='Shared',
  full_name='protobuf.distributions.DirichletDiscrete.Shared',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='alphas', full_name='protobuf.distributions.DirichletDiscrete.Shared.alphas', index=0,
      number=1, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=440,
  serialized_end=464,
)

_DIRICHLETDISCRETE_GRIDPRIOR = descriptor.Descriptor(
  name='GridPrior',
  full_name='protobuf.distributions.DirichletDiscrete.GridPrior',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='alpha', full_name='protobuf.distributions.DirichletDiscrete.GridPrior.alpha', index=0,
      number=1, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=338,
  serialized_end=364,
)

_DIRICHLETDISCRETE_GROUP = descriptor.Descriptor(
  name='Group',
  full_name='protobuf.distributions.DirichletDiscrete.Group',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='counts', full_name='protobuf.distributions.DirichletDiscrete.Group.counts', index=0,
      number=1, type=4, cpp_type=4, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=494,
  serialized_end=517,
)

_DIRICHLETDISCRETE = descriptor.Descriptor(
  name='DirichletDiscrete',
  full_name='protobuf.distributions.DirichletDiscrete',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[_DIRICHLETDISCRETE_SHARED, _DIRICHLETDISCRETE_GRIDPRIOR, _DIRICHLETDISCRETE_GROUP, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=419,
  serialized_end=517,
)


_DIRICHLETPROCESSDISCRETE_SHARED = descriptor.Descriptor(
  name='Shared',
  full_name='protobuf.distributions.DirichletProcessDiscrete.Shared',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='gamma', full_name='protobuf.distributions.DirichletProcessDiscrete.Shared.gamma', index=0,
      number=1, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='alpha', full_name='protobuf.distributions.DirichletProcessDiscrete.Shared.alpha', index=1,
      number=2, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='values', full_name='protobuf.distributions.DirichletProcessDiscrete.Shared.values', index=2,
      number=3, type=13, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='betas', full_name='protobuf.distributions.DirichletProcessDiscrete.Shared.betas', index=3,
      number=4, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='counts', full_name='protobuf.distributions.DirichletProcessDiscrete.Shared.counts', index=4,
      number=5, type=4, cpp_type=4, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=548,
  serialized_end=633,
)

_DIRICHLETPROCESSDISCRETE_GRIDPRIOR = descriptor.Descriptor(
  name='GridPrior',
  full_name='protobuf.distributions.DirichletProcessDiscrete.GridPrior',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='gamma', full_name='protobuf.distributions.DirichletProcessDiscrete.GridPrior.gamma', index=0,
      number=1, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='alpha', full_name='protobuf.distributions.DirichletProcessDiscrete.GridPrior.alpha', index=1,
      number=2, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=635,
  serialized_end=676,
)

_DIRICHLETPROCESSDISCRETE_GROUP = descriptor.Descriptor(
  name='Group',
  full_name='protobuf.distributions.DirichletProcessDiscrete.Group',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='keys', full_name='protobuf.distributions.DirichletProcessDiscrete.Group.keys', index=0,
      number=1, type=13, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='values', full_name='protobuf.distributions.DirichletProcessDiscrete.Group.values', index=1,
      number=2, type=4, cpp_type=4, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=678,
  serialized_end=715,
)

_DIRICHLETPROCESSDISCRETE = descriptor.Descriptor(
  name='DirichletProcessDiscrete',
  full_name='protobuf.distributions.DirichletProcessDiscrete',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[_DIRICHLETPROCESSDISCRETE_SHARED, _DIRICHLETPROCESSDISCRETE_GRIDPRIOR, _DIRICHLETPROCESSDISCRETE_GROUP, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=520,
  serialized_end=715,
)


_PITMANYORPROCESSDISCRETE_SHARED = descriptor.Descriptor(
  name='Shared',
  full_name='protobuf.distributions.PitmanYorProcessDiscrete.Shared',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='alpha', full_name='protobuf.distributions.PitmanYorProcessDiscrete.Shared.alpha', index=0,
      number=1, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='d', full_name='protobuf.distributions.PitmanYorProcessDiscrete.Shared.d', index=1,
      number=2, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='counts', full_name='protobuf.distributions.PitmanYorProcessDiscrete.Shared.counts', index=2,
      number=3, type=4, cpp_type=4, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=746,
  serialized_end=796,
)

_PITMANYORPROCESSDISCRETE_GRIDPRIOR = descriptor.Descriptor(
  name='GridPrior',
  full_name='protobuf.distributions.PitmanYorProcessDiscrete.GridPrior',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='alpha_d', full_name='protobuf.distributions.PitmanYorProcessDiscrete.GridPrior.alpha_d', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=798,
  serialized_end=872,
)

_PITMANYORPROCESSDISCRETE_GROUP = descriptor.Descriptor(
  name='Group',
  full_name='protobuf.distributions.PitmanYorProcessDiscrete.Group',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='keys', full_name='protobuf.distributions.PitmanYorProcessDiscrete.Group.keys', index=0,
      number=1, type=13, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='values', full_name='protobuf.distributions.PitmanYorProcessDiscrete.Group.values', index=1,
      number=2, type=4, cpp_type=4, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=678,
  serialized_end=715,
)

_PITMANYORPROCESSDISCRETE = descriptor.Descriptor(
  name='PitmanYorProcessDiscrete',
  full_name='protobuf.distributions.PitmanYorProcessDiscrete',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[_PITMANYORPROCESSDISCRETE_SHARED, _PITMANYORPROCESSDISCRETE_GRIDPRIOR, _PITMANYORPROCESSDISCRETE_GROUP, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=718,
  serialized_end=911,
)


_GAMMAPOISSON_SHARED = descriptor.Descriptor(
  name='Shared',
  full_name='protobuf.distributions.GammaPoisson.Shared',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='alpha', full_name='protobuf.distributions.GammaPoisson.Shared.alpha', index=0,
      number=1, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='inv_beta', full_name='protobuf.distributions.GammaPoisson.Shared.inv_beta', index=1,
      number=2, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=930,
  serialized_end=971,
)

_GAMMAPOISSON_GRIDPRIOR = descriptor.Descriptor(
  name='GridPrior',
  full_name='protobuf.distributions.GammaPoisson.GridPrior',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='alpha', full_name='protobuf.distributions.GammaPoisson.GridPrior.alpha', index=0,
      number=1, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='inv_beta', full_name='protobuf.distributions.GammaPoisson.GridPrior.inv_beta', index=1,
      number=2, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=973,
  serialized_end=1017,
)

_GAMMAPOISSON_GROUP = descriptor.Descriptor(
  name='Group',
  full_name='protobuf.distributions.GammaPoisson.Group',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='count', full_name='protobuf.distributions.GammaPoisson.Group.count', index=0,
      number=1, type=4, cpp_type=4, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='sum', full_name='protobuf.distributions.GammaPoisson.Group.sum', index=1,
      number=2, type=4, cpp_type=4, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='log_prod', full_name='protobuf.distributions.GammaPoisson.Group.log_prod', index=2,
      number=3, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1019,
  serialized_end=1072,
)

_GAMMAPOISSON = descriptor.Descriptor(
  name='GammaPoisson',
  full_name='protobuf.distributions.GammaPoisson',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[_GAMMAPOISSON_SHARED, _GAMMAPOISSON_GRIDPRIOR, _GAMMAPOISSON_GROUP, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=914,
  serialized_end=1072,
)


_BETANEGATIVEBINOMIAL_SHARED = descriptor.Descriptor(
  name='Shared',
  full_name='protobuf.distributions.BetaNegativeBinomial.Shared',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='alpha', full_name='protobuf.distributions.BetaNegativeBinomial.Shared.alpha', index=0,
      number=1, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='beta', full_name='protobuf.distributions.BetaNegativeBinomial.Shared.beta', index=1,
      number=2, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='r', full_name='protobuf.distributions.BetaNegativeBinomial.Shared.r', index=2,
      number=3, type=4, cpp_type=4, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1099,
  serialized_end=1147,
)

_BETANEGATIVEBINOMIAL_GRIDPRIOR = descriptor.Descriptor(
  name='GridPrior',
  full_name='protobuf.distributions.BetaNegativeBinomial.GridPrior',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='alpha', full_name='protobuf.distributions.BetaNegativeBinomial.GridPrior.alpha', index=0,
      number=1, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='beta', full_name='protobuf.distributions.BetaNegativeBinomial.GridPrior.beta', index=1,
      number=2, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='r', full_name='protobuf.distributions.BetaNegativeBinomial.GridPrior.r', index=2,
      number=3, type=4, cpp_type=4, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1149,
  serialized_end=1200,
)

_BETANEGATIVEBINOMIAL_GROUP = descriptor.Descriptor(
  name='Group',
  full_name='protobuf.distributions.BetaNegativeBinomial.Group',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='count', full_name='protobuf.distributions.BetaNegativeBinomial.Group.count', index=0,
      number=1, type=4, cpp_type=4, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='sum', full_name='protobuf.distributions.BetaNegativeBinomial.Group.sum', index=1,
      number=2, type=4, cpp_type=4, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1019,
  serialized_end=1054,
)

_BETANEGATIVEBINOMIAL = descriptor.Descriptor(
  name='BetaNegativeBinomial',
  full_name='protobuf.distributions.BetaNegativeBinomial',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[_BETANEGATIVEBINOMIAL_SHARED, _BETANEGATIVEBINOMIAL_GRIDPRIOR, _BETANEGATIVEBINOMIAL_GROUP, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1075,
  serialized_end=1237,
)


_NORMALINVERSECHISQ_SHARED = descriptor.Descriptor(
  name='Shared',
  full_name='protobuf.distributions.NormalInverseChiSq.Shared',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='mu', full_name='protobuf.distributions.NormalInverseChiSq.Shared.mu', index=0,
      number=1, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='kappa', full_name='protobuf.distributions.NormalInverseChiSq.Shared.kappa', index=1,
      number=2, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='sigmasq', full_name='protobuf.distributions.NormalInverseChiSq.Shared.sigmasq', index=2,
      number=3, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='nu', full_name='protobuf.distributions.NormalInverseChiSq.Shared.nu', index=3,
      number=4, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1262,
  serialized_end=1326,
)

_NORMALINVERSECHISQ_GRIDPRIOR = descriptor.Descriptor(
  name='GridPrior',
  full_name='protobuf.distributions.NormalInverseChiSq.GridPrior',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='mu', full_name='protobuf.distributions.NormalInverseChiSq.GridPrior.mu', index=0,
      number=1, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='kappa', full_name='protobuf.distributions.NormalInverseChiSq.GridPrior.kappa', index=1,
      number=2, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='sigmasq', full_name='protobuf.distributions.NormalInverseChiSq.GridPrior.sigmasq', index=2,
      number=3, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='nu', full_name='protobuf.distributions.NormalInverseChiSq.GridPrior.nu', index=3,
      number=4, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1328,
  serialized_end=1395,
)

_NORMALINVERSECHISQ_GROUP = descriptor.Descriptor(
  name='Group',
  full_name='protobuf.distributions.NormalInverseChiSq.Group',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    descriptor.FieldDescriptor(
      name='count', full_name='protobuf.distributions.NormalInverseChiSq.Group.count', index=0,
      number=1, type=4, cpp_type=4, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='mean', full_name='protobuf.distributions.NormalInverseChiSq.Group.mean', index=1,
      number=2, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    descriptor.FieldDescriptor(
      name='count_times_variance', full_name='protobuf.distributions.NormalInverseChiSq.Group.count_times_variance', index=2,
      number=3, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1397,
  serialized_end=1463,
)

_NORMALINVERSECHISQ = descriptor.Descriptor(
  name='NormalInverseChiSq',
  full_name='protobuf.distributions.NormalInverseChiSq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[_NORMALINVERSECHISQ_SHARED, _NORMALINVERSECHISQ_GRIDPRIOR, _NORMALINVERSECHISQ_GROUP, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=1240,
  serialized_end=1463,
)

_CLUSTERING_PITMANYOR.containing_type = _CLUSTERING;
_CLUSTERING_LOWENTROPY.containing_type = _CLUSTERING;
_CLUSTERING.fields_by_name['pitman_yor'].message_type = _CLUSTERING_PITMANYOR
_CLUSTERING.fields_by_name['low_entropy'].message_type = _CLUSTERING_LOWENTROPY
_BETABERNOULLI_SHARED.containing_type = _BETABERNOULLI;
_BETABERNOULLI_GRIDPRIOR.containing_type = _BETABERNOULLI;
_BETABERNOULLI_GROUP.containing_type = _BETABERNOULLI;
_DIRICHLETDISCRETE_SHARED.containing_type = _DIRICHLETDISCRETE;
_DIRICHLETDISCRETE_GRIDPRIOR.containing_type = _DIRICHLETDISCRETE;
_DIRICHLETDISCRETE_GROUP.containing_type = _DIRICHLETDISCRETE;
_DIRICHLETPROCESSDISCRETE_SHARED.containing_type = _DIRICHLETPROCESSDISCRETE;
_DIRICHLETPROCESSDISCRETE_GRIDPRIOR.containing_type = _DIRICHLETPROCESSDISCRETE;
_DIRICHLETPROCESSDISCRETE_GROUP.containing_type = _DIRICHLETPROCESSDISCRETE;
_PITMANYORPROCESSDISCRETE_SHARED.containing_type = _PITMANYORPROCESSDISCRETE;
_PITMANYORPROCESSDISCRETE_GRIDPRIOR.fields_by_name['alpha_d'].message_type = _CLUSTERING_PITMANYOR
_PITMANYORPROCESSDISCRETE_GRIDPRIOR.containing_type = _PITMANYORPROCESSDISCRETE;
_PITMANYORPROCESSDISCRETE_GROUP.containing_type = _PITMANYORPROCESSDISCRETE;
_GAMMAPOISSON_SHARED.containing_type = _GAMMAPOISSON;
_GAMMAPOISSON_GRIDPRIOR.containing_type = _GAMMAPOISSON;
_GAMMAPOISSON_GROUP.containing_type = _GAMMAPOISSON;
_BETANEGATIVEBINOMIAL_SHARED.containing_type = _BETANEGATIVEBINOMIAL;
_BETANEGATIVEBINOMIAL_GRIDPRIOR.containing_type = _BETANEGATIVEBINOMIAL;
_BETANEGATIVEBINOMIAL_GROUP.containing_type = _BETANEGATIVEBINOMIAL;
_NORMALINVERSECHISQ_SHARED.containing_type = _NORMALINVERSECHISQ;
_NORMALINVERSECHISQ_GRIDPRIOR.containing_type = _NORMALINVERSECHISQ;
_NORMALINVERSECHISQ_GROUP.containing_type = _NORMALINVERSECHISQ;
DESCRIPTOR.message_types_by_name['Clustering'] = _CLUSTERING
DESCRIPTOR.message_types_by_name['BetaBernoulli'] = _BETABERNOULLI
DESCRIPTOR.message_types_by_name['DirichletDiscrete'] = _DIRICHLETDISCRETE
DESCRIPTOR.message_types_by_name['DirichletProcessDiscrete'] = _DIRICHLETPROCESSDISCRETE
DESCRIPTOR.message_types_by_name['PitmanYorProcessDiscrete'] = _PITMANYORPROCESSDISCRETE
DESCRIPTOR.message_types_by_name['GammaPoisson'] = _GAMMAPOISSON
DESCRIPTOR.message_types_by_name['BetaNegativeBinomial'] = _BETANEGATIVEBINOMIAL
DESCRIPTOR.message_types_by_name['NormalInverseChiSq'] = _NORMALINVERSECHISQ

class Clustering(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  
  class PitmanYor(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _CLUSTERING_PITMANYOR
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.Clustering.PitmanYor)
  
  class LowEntropy(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _CLUSTERING_LOWENTROPY
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.Clustering.LowEntropy)
  DESCRIPTOR = _CLUSTERING
  
  # @@protoc_insertion_point(class_scope:protobuf.distributions.Clustering)

class BetaBernoulli(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  
  class Shared(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _BETABERNOULLI_SHARED
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.BetaBernoulli.Shared)
  
  class GridPrior(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _BETABERNOULLI_GRIDPRIOR
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.BetaBernoulli.GridPrior)
  
  class Group(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _BETABERNOULLI_GROUP
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.BetaBernoulli.Group)
  DESCRIPTOR = _BETABERNOULLI
  
  # @@protoc_insertion_point(class_scope:protobuf.distributions.BetaBernoulli)

class DirichletDiscrete(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  
  class Shared(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _DIRICHLETDISCRETE_SHARED
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.DirichletDiscrete.Shared)
  
  class GridPrior(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _DIRICHLETDISCRETE_GRIDPRIOR
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.DirichletDiscrete.GridPrior)
  
  class Group(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _DIRICHLETDISCRETE_GROUP
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.DirichletDiscrete.Group)
  DESCRIPTOR = _DIRICHLETDISCRETE
  
  # @@protoc_insertion_point(class_scope:protobuf.distributions.DirichletDiscrete)

class DirichletProcessDiscrete(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  
  class Shared(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _DIRICHLETPROCESSDISCRETE_SHARED
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.DirichletProcessDiscrete.Shared)
  
  class GridPrior(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _DIRICHLETPROCESSDISCRETE_GRIDPRIOR
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.DirichletProcessDiscrete.GridPrior)
  
  class Group(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _DIRICHLETPROCESSDISCRETE_GROUP
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.DirichletProcessDiscrete.Group)
  DESCRIPTOR = _DIRICHLETPROCESSDISCRETE
  
  # @@protoc_insertion_point(class_scope:protobuf.distributions.DirichletProcessDiscrete)

class PitmanYorProcessDiscrete(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  
  class Shared(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _PITMANYORPROCESSDISCRETE_SHARED
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.PitmanYorProcessDiscrete.Shared)
  
  class GridPrior(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _PITMANYORPROCESSDISCRETE_GRIDPRIOR
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.PitmanYorProcessDiscrete.GridPrior)
  
  class Group(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _PITMANYORPROCESSDISCRETE_GROUP
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.PitmanYorProcessDiscrete.Group)
  DESCRIPTOR = _PITMANYORPROCESSDISCRETE
  
  # @@protoc_insertion_point(class_scope:protobuf.distributions.PitmanYorProcessDiscrete)

class GammaPoisson(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  
  class Shared(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _GAMMAPOISSON_SHARED
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.GammaPoisson.Shared)
  
  class GridPrior(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _GAMMAPOISSON_GRIDPRIOR
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.GammaPoisson.GridPrior)
  
  class Group(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _GAMMAPOISSON_GROUP
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.GammaPoisson.Group)
  DESCRIPTOR = _GAMMAPOISSON
  
  # @@protoc_insertion_point(class_scope:protobuf.distributions.GammaPoisson)

class BetaNegativeBinomial(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  
  class Shared(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _BETANEGATIVEBINOMIAL_SHARED
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.BetaNegativeBinomial.Shared)
  
  class GridPrior(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _BETANEGATIVEBINOMIAL_GRIDPRIOR
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.BetaNegativeBinomial.GridPrior)
  
  class Group(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _BETANEGATIVEBINOMIAL_GROUP
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.BetaNegativeBinomial.Group)
  DESCRIPTOR = _BETANEGATIVEBINOMIAL
  
  # @@protoc_insertion_point(class_scope:protobuf.distributions.BetaNegativeBinomial)

class NormalInverseChiSq(message.Message):
  __metaclass__ = reflection.GeneratedProtocolMessageType
  
  class Shared(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _NORMALINVERSECHISQ_SHARED
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.NormalInverseChiSq.Shared)
  
  class GridPrior(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _NORMALINVERSECHISQ_GRIDPRIOR
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.NormalInverseChiSq.GridPrior)
  
  class Group(message.Message):
    __metaclass__ = reflection.GeneratedProtocolMessageType
    DESCRIPTOR = _NORMALINVERSECHISQ_GROUP
    
    # @@protoc_insertion_point(class_scope:protobuf.distributions.NormalInverseChiSq.Group)
  DESCRIPTOR = _NORMALINVERSECHISQ
  
  # @@protoc_insertion_point(class_scope:protobuf.distributions.NormalInverseChiSq)

# @@protoc_insertion_point(module_scope)



# Copyright (c) 2014, Salesforce.com, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# - Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# - Neither the name of Salesforce.com nor the names of its contributors
#   may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
from scipy.special import betaln, binom
from matplotlib import pyplot


def binomln(n, k):
    return np.log(binom(n, k))


def simone(alpha, beta, n, ITERS):
    ks = []
    for i in range(ITERS):
        theta = np.random.beta(alpha, beta)
        ks.append(np.random.binomial(n, theta))
    return np.array(ks)


def simtwo(alpha, beta, n, ITERS):
    counts = np.zeros((ITERS, 2), dtype=np.uint32)
    thetas = np.random.beta(alpha, beta, size=ITERS)
    for i in range(ITERS):
        if i % 1000000 == 0:
            print i

        counts[i] = np.random.binomial(n, thetas[i], size=2)
    return counts


class Hypers(object):

    def __init__(self):
        self.alpha = 0.0
        self.beta = 0.0
        self.N = 0

    def __str__(self):
        return "alpha=%f, beta=%f, N=%d" % (self.alpha,
                                            self.beta,
                                            self.N)


class SuffStats(object):

    def __init__(self):
        self.heads = 0
        self.tails = 0
        self.binomln_accum = 0
        self.dpcount = 0

    def __str__(self):
        return "heads=%d, tails=%d" % (self.heads, self.tails)


def add_dp(k, hps, suffs):
    suffs.heads += k
    suffs.tails += hps.N - k
    suffs.binomln_accum += binomln(hps.N, k)

    suffs.dpcount += 1


def compute_post_pred(k, hps, suffs):

    a = hps.alpha + suffs.heads
    b = hps.beta + suffs.tails

    return binomln(hps.N, k) + betaln(a + k, b + hps.N - k) - betaln(a, b)


def compute_total_likelihood(hps, suffs):

    a = hps.alpha + suffs.heads
    b = hps.beta + suffs.tails

    r = betaln(a, b) - betaln(hps.alpha, hps.beta)

    r += + suffs.binomln_accum

    return r


def two_d_hist(alpha, beta, N, ITERS):

    r = simtwo(alpha, beta, N, ITERS)
    h, b1, b2 = np.histogram2d(r[:, 0], r[:, 1], bins=N + 1)
    h[:, :] += 1
    pyplot.imshow(h)
    pyplot.show()

    logp = np.log(h.astype(float) / (ITERS))

    return logp


def test_post_pred_equal_likelihood():
    hps = Hypers()
    hps.N = 10

    for alpha in [5.0]:
        for beta in [5.0]:
            for ks in [[9, 1], [1, 9]]:
                print "-" * 60
                hps.alpha = alpha
                hps.beta = beta

                ss = SuffStats()

                pred_score = 0.0
                for ki, k in enumerate(ks):
                    pred_score += compute_post_pred(k, hps, ss)
                    add_dp(k, hps, ss)

                total_score = compute_total_likelihood(hps, ss)
                res = two_d_hist(alpha, beta, hps.N, 10000000)
                print res
                print res.shape
                emp_score = res[ks[0], ks[1]]
                d1 = abs(pred_score - emp_score)
                d2 = abs(emp_score - total_score)
                d3 = abs(total_score - pred_score)
                print "alpha=", alpha, "beta=", beta, "ks =", ks
                print pred_score, total_score, emp_score
                print "error=", d1 + d2 + d3



# Copyright (c) 2014, Salesforce.com, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# - Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# - Neither the name of Salesforce.com nor the names of its contributors
#   may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy
cimport numpy

from distributions.rng_cc cimport rng_t
from distributions.global_rng cimport get_rng

cdef extern from 'rng_cc.hpp' namespace 'std_wrapper':
    cdef void std_rng_seed(rng_t & rng, unsigned long seed)
    cdef double std_random_normal(rng_t & rng, double mu, double sigmasq)
    cdef double std_random_chisq(rng_t & rng, double nu)
    cdef double std_random_gamma(rng_t & rng, double alpha, double beta)
    cdef int std_random_poisson(rng_t & rng, double mu)
    cdef int std_random_categorical(rng_t & rng, size_t dim, double ps[])
    cdef void std_random_dirichlet(
            rng_t & rng,
            size_t dim,
            double alphas[],
            double thetas[])


cpdef int random():
    return get_rng()[0].sample()


cpdef seed(unsigned long s):
    std_rng_seed(get_rng()[0], s)


cpdef double sample_normal(double mu, double sigmasq):
    return std_random_normal(get_rng()[0], mu, sigmasq)


cpdef double sample_chisq(double nu):
    return std_random_chisq(get_rng()[0], nu)


cpdef sample_gamma(double a, double b):
    return std_random_gamma(get_rng()[0], a, b)


cpdef sample_poisson(double mu):
    return std_random_poisson(get_rng()[0], mu)


cdef sample_discrete(size_t dim, double ps[]):
    return std_random_categorical(get_rng()[0], dim, ps)


cdef sample_dirichlet(size_t dim, double alphas[], double thetas[]):
    std_random_dirichlet(get_rng()[0], dim, alphas, thetas)


#def sample_discrete(numpy.ndarray[double, ndim=1] ps):
#    cdef int dim = ps.shape[0]
#    return std_random_categorical(dim, &ps[0])
#
#
#def sample_dirichlet(
#        numpy.ndarray[double, ndim=1] alphas,
#        numpy.ndarray[double, ndim=1] thetas):
#    cdef int dim = alphas.shape[0]
#    std_random_dirichlet(dim, &alphas[0], &thetas[0])



# Copyright (c) 2014, Salesforce.com, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# - Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# - Neither the name of Salesforce.com nor the names of its contributors
#   may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

cimport _gp
import _gp

from distributions.mixins import SharedMixin, GroupIoMixin, SharedIoMixin


NAME = 'GammaPoisson'
EXAMPLES = [
    {
        'shared': {'alpha': 1., 'inv_beta': 1.},
        'values': [0, 1, 2, 3, 4, 5, 6, 1, 2, 3, 4, 2, 3],
    },
]
Value = int


cdef class _Shared(_gp.Shared):
    def load(self, raw):
        self.ptr.alpha = raw['alpha']
        self.ptr.inv_beta = raw['inv_beta']

    def dump(self):
        return {
            'alpha': self.ptr.alpha,
            'inv_beta': self.ptr.inv_beta,
        }

    def load_protobuf(self, message):
        self.ptr.alpha = message.alpha
        self.ptr.inv_beta = message.inv_beta

    def dump_protobuf(self, message):
        message.alpha = self.ptr.alpha
        message.inv_beta = self.ptr.inv_beta


class Shared(_Shared, SharedMixin, SharedIoMixin):
    pass


cdef class _Group(_gp.Group):
    def load(self, raw):
        self.ptr.count = raw['count']
        self.ptr.sum = raw['sum']
        self.ptr.log_prod = raw['log_prod']

    def dump(self):
        return {
            'count': self.ptr.count,
            'sum': self.ptr.sum,
            'log_prod': self.ptr.log_prod,
        }


class Group(_Group, GroupIoMixin):
    pass


class Sampler(_gp.Sampler):
    pass


Mixture = _gp.Mixture
sample_group = _gp.sample_group



# Copyright (c) 2014, Salesforce.com, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# - Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# - Neither the name of Salesforce.com nor the names of its contributors
#   may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import re
import sys
import numpy

if len(sys.argv) >= 2 and sys.argv[1] == 'bdist_wheel':
    # bdist_wheel needs setuptools
    import setuptools
    assert setuptools  # satisfy pyflakes

from distutils.core import setup, Extension
from distutils.version import LooseVersion

try:
    from Cython.Build import cythonize
    from cython import __version__ as cython_version
    min_cython_version = '0.20.1'
    if LooseVersion(cython_version) < LooseVersion(min_cython_version):
        raise ValueError(
            'cython support requires cython>={}'.format(min_cython_version))
    cython = True
except ImportError:
    cython = False


clang = False
if sys.platform.lower().startswith('darwin'):
    clang = True


include_dirs = ['include', 'distributions']
include_dirs.append(numpy.get_include())


extra_compile_args = [
    '-DDIST_DEBUG_LEVEL=3',
    '-DDIST_THROW_ON_ERROR=1',
]
if clang:
    extra_compile_args.extend([
        '-mmacosx-version-min=10.7',  # for anaconda
        '-std=c++0x',
        '-stdlib=libc++',
    ])
else:
    extra_compile_args.extend([
        '-std=c++0x',
        '-Wall',
        '-Werror',
        '-Wno-error=cpp',  # avoid #warning "Using deprecated NumPy API,..."
        '-Wno-unused-function',
        '-Wno-sign-compare',
        '-Wno-strict-aliasing',
        '-O3',
        '-ffast-math',
        '-funsafe-math-optimizations',
        #'-fno-trapping-math',
        #'-ffinite-math-only',
        #'-fvect-cost-model',
        '-mfpmath=sse',
        '-msse4.1',
        #'-mavx',
        #'-mrecip',
        #'-march=native',
        #'-DNPY_NO_DEPRECATED_API=NPY_1_7_API_VERSION',
    ])


use_libdistributions = 'PYDISTRIBUTIONS_USE_LIB' in os.environ
use_protobuf = 'DISTRIBUTIONS_USE_PROTOBUF' in os.environ


def make_extension(name):
    module = 'distributions.' + name
    sources = [
        '{}.{}'.format(module.replace('.', '/'), 'pyx' if cython else 'cpp')
    ]
    libraries = ['m']
    if use_protobuf:
        libraries.append('protobuf')
    if name.startswith('lp'):
        if use_libdistributions:
            libraries = ['distributions_shared'] + libraries
        else:
            sources += [
                'src/common.cc',
                'src/special.cc',
                'src/random.cc',
                'src/vector_math.cc',
            ]
            if name == 'lp.clustering':
                sources.append('src/clustering.cc')
            elif name in ('lp.models.nich', 'lp.models._nich'):
                sources.append('src/models/nich.cc')
            elif name in ('lp.models.gp', 'lp.models._gp'):
                sources.append('src/models/gp.cc')
    return Extension(
        module,
        sources=sources,
        language='c++',
        include_dirs=include_dirs,
        libraries=libraries,
        extra_compile_args=extra_compile_args,
    )


def make_extensions(names):
    return [make_extension(name) for name in names]


hp_extensions = make_extensions([
    'has_cython',
    'rng_cc',
    'global_rng',
    'hp.special',
    'hp.random',
    'hp.models.dd',
    'hp.models.gp',
    'hp.models.nich',
])


lp_extensions = make_extensions([
    'lp.special',
    'lp.random',
    'lp.vector',
    'lp.models.bb',
    'lp.models._bb',
    'lp.models.dd',
    'lp.models._dd',
    'lp.models.dpd',
    'lp.models._dpd',
    'lp.models.gp',
    'lp.models._gp',
    'lp.models.bnb',
    'lp.models._bnb',
    'lp.models.nich',
    'lp.models._nich',
    'lp.clustering',
    'lp.mixture',
])


if cython:
    ext_modules = cythonize(hp_extensions + lp_extensions)
else:
    ext_modules = hp_extensions + lp_extensions


version = None
with open(os.path.join('distributions', '__init__.py')) as f:
    for line in f:
        if re.match("__version__ = '\S+'$", line):
            version = line.split()[-1].strip("'")
assert version, 'could not determine version'


with open('README.md') as f:
    long_description = f.read()


config = {
    'version': version,
    'name': 'distributions',
    'description': 'Primitives for Bayesian MCMC inference',
    'long_description': long_description,
    'url': 'https://github.com/forcedotcom/distributions',
    'author': 'Jonathan Glidden, Eric Jonas, Fritz Obermeyer, Cap Petschulat',
    'maintainer': 'Fritz Obermeyer',
    'maintainer_email': 'fritz.obermeyer@gmail.com',
    'license': 'Revised BSD',
    'packages': [
        'distributions',
        'distributions.dbg',
        'distributions.dbg.models',
        'distributions.hp',
        'distributions.hp.models',
        'distributions.lp',
        'distributions.lp.models',
        'distributions.io',
        'distributions.tests',
    ],
    'ext_modules': ext_modules,
}


setup(**config)



# Copyright (c) 2014, Salesforce.com, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# - Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# - Neither the name of Salesforce.com nor the names of its contributors
#   may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import shutil
import numpy
import scipy
import scipy.misc
import scipy.ndimage
from distributions.dbg.random import sample_discrete, sample_discrete_log
from distributions.lp.models import nich
from distributions.lp.clustering import PitmanYor
from distributions.lp.mixture import MixtureIdTracker
from distributions.io.stream import json_stream_load, json_stream_dump
from multiprocessing import Process
import parsable
parsable = parsable.Parsable()


ROOT = os.path.dirname(os.path.abspath(__file__))
DATA = os.path.join(ROOT, 'data')
RESULTS = os.path.join(ROOT, 'results')
SAMPLES = os.path.join(DATA, 'samples.json.gz')
IMAGE = scipy.misc.imread(os.path.join(ROOT, 'fox.png'))
SAMPLE_COUNT = 10000
PASSES = 10
EMPTY_GROUP_COUNT = 10


for dirname in [DATA, RESULTS]:
    if not os.path.exists(dirname):
        os.makedirs(dirname)


class ImageModel(object):
    def __init__(self):
        self.clustering = PitmanYor.from_dict({
            'alpha': 100.0,
            'd': 0.1,
        })
        self.feature = nich.Shared.from_dict({
            'mu': 0.0,
            'kappa': 0.1,
            'sigmasq': 0.01,
            'nu': 1.0,
        })

    class Mixture(object):
        def __init__(self):
            self.clustering = PitmanYor.Mixture()
            self.feature_x = nich.Mixture()
            self.feature_y = nich.Mixture()
            self.id_tracker = MixtureIdTracker()

        def __len__(self):
            return len(self.clustering)

        def init(self, model, empty_group_count=EMPTY_GROUP_COUNT):
            assert empty_group_count >= 1
            counts = [0] * empty_group_count
            self.clustering.init(model.clustering, counts)
            assert len(self.clustering) == len(counts)
            self.id_tracker.init(len(counts))

            self.feature_x.clear()
            self.feature_y.clear()
            for _ in xrange(empty_group_count):
                self.feature_x.add_group(model.feature)
                self.feature_y.add_group(model.feature)
            self.feature_x.init(model.feature)
            self.feature_y.init(model.feature)

        def score_value(self, model, xy, scores):
            x, y = xy
            self.clustering.score_value(model.clustering, scores)
            self.feature_x.score_value(model.feature, x, scores)
            self.feature_y.score_value(model.feature, y, scores)

        def add_value(self, model, groupid, xy):
            x, y = xy
            group_added = self.clustering.add_value(model.clustering, groupid)
            self.feature_x.add_value(model.feature, groupid, x)
            self.feature_y.add_value(model.feature, groupid, y)
            if group_added:
                self.feature_x.add_group(model.feature)
                self.feature_y.add_group(model.feature)
                self.id_tracker.add_group()

        def remove_value(self, model, groupid, xy):
            x, y = xy
            group_removeed = self.clustering.remove_value(
                model.clustering,
                groupid)
            self.feature_x.remove_value(model.feature, groupid, x)
            self.feature_y.remove_value(model.feature, groupid, y)
            if group_removeed:
                self.feature_x.remove_group(model.feature, groupid)
                self.feature_y.remove_group(model.feature, groupid)
                self.id_tracker.remove_group(groupid)


def sample_from_image(image, sample_count):
    image = -1.0 * image
    image -= image.min()
    x_pmf = image.sum(axis=1)
    y_pmfs = image.copy()
    for y_pmf in y_pmfs:
        y_pmf /= (y_pmf.sum() + 1e-8)

    x_scale = 2.0 / (image.shape[0] - 1)
    y_scale = 2.0 / (image.shape[1] - 1)

    for _ in xrange(sample_count):
        x = sample_discrete(x_pmf)
        y = sample_discrete(y_pmfs[x])
        yield (x * x_scale - 1.0, y * y_scale - 1.0)


def synthesize_image(model, mixture):
    width, height = IMAGE.shape
    image = numpy.zeros((width, height))
    scores = numpy.zeros(len(mixture), dtype=numpy.float32)
    x_scale = 2.0 / (width - 1)
    y_scale = 2.0 / (height - 1)
    for x in xrange(width):
        for y in xrange(height):
            xy = (x * x_scale - 1.0, y * y_scale - 1.0)
            mixture.score_value(model, xy, scores)
            prob = numpy.exp(scores, out=scores).sum()
            image[x, y] = prob

    image /= image.max()
    image -= 1.0
    image *= -255
    return image.astype(numpy.uint8)


def visualize_dataset(samples):
    width, height = IMAGE.shape
    x_scale = 2.0 / (width - 1)
    y_scale = 2.0 / (height - 1)
    image = numpy.zeros((width, height))
    for x, y in samples:
        x = int(round((x + 1.0) / x_scale))
        y = int(round((y + 1.0) / y_scale))
        image[x, y] += 1
    image = scipy.ndimage.gaussian_filter(image, sigma=1)
    image *= -255.0 / image.max()
    image -= image.min()
    return image.astype(numpy.uint8)


@parsable.command
def create_dataset(sample_count=SAMPLE_COUNT):
    '''
    Extract dataset from image.
    '''
    scipy.misc.imsave(os.path.join(RESULTS, 'original.png'), IMAGE)
    print 'sampling {} points from image'.format(sample_count)
    samples = sample_from_image(IMAGE, sample_count)
    json_stream_dump(samples, SAMPLES)
    image = visualize_dataset(json_stream_load(SAMPLES))
    scipy.misc.imsave(os.path.join(RESULTS, 'samples.png'), image)


@parsable.command
def compress_sequential():
    '''
    Compress image via sequential initialization.
    '''
    assert os.path.exists(SAMPLES), 'first create dataset'
    print 'sequential start'
    model = ImageModel()
    mixture = ImageModel.Mixture()
    mixture.init(model)
    scores = numpy.zeros(1, dtype=numpy.float32)

    for xy in json_stream_load(SAMPLES):
        scores.resize(len(mixture))
        mixture.score_value(model, xy, scores)
        groupid = sample_discrete_log(scores)
        mixture.add_value(model, groupid, xy)

    print 'sequential found {} components'.format(len(mixture))
    image = synthesize_image(model, mixture)
    scipy.misc.imsave(os.path.join(RESULTS, 'sequential.png'), image)


@parsable.command
def compress_gibbs(passes=PASSES):
    '''
    Compress image via gibbs sampling.
    '''
    assert passes >= 0
    assert os.path.exists(SAMPLES), 'first create dataset'
    print 'prior+gibbs start {} passes'.format(passes)
    model = ImageModel()
    mixture = ImageModel.Mixture()
    mixture.init(model)
    scores = numpy.zeros(1, dtype=numpy.float32)
    assignments = {}

    for i, xy in enumerate(json_stream_load(SAMPLES)):
        scores.resize(len(mixture))
        mixture.clustering.score_value(model.clustering, scores)
        groupid = sample_discrete_log(scores)
        mixture.add_value(model, groupid, xy)
        assignments[i] = mixture.id_tracker.packed_to_global(groupid)

    print 'prior+gibbs init with {} components'.format(len(mixture))

    for _ in xrange(passes):
        for i, xy in enumerate(json_stream_load(SAMPLES)):
            groupid = mixture.id_tracker.global_to_packed(assignments[i])
            mixture.remove_value(model, groupid, xy)
            scores.resize(len(mixture))
            mixture.score_value(model, xy, scores)
            groupid = sample_discrete_log(scores)
            mixture.add_value(model, groupid, xy)
            assignments[i] = mixture.id_tracker.packed_to_global(groupid)

    print 'prior+gibbs found {} components'.format(len(mixture))
    image = synthesize_image(model, mixture)
    scipy.misc.imsave(os.path.join(RESULTS, 'prior_gibbs.png'), image)


@parsable.command
def compress_seq_gibbs(passes=PASSES):
    '''
    Compress image via sequentiall-initialized gibbs sampling.
    '''
    assert passes >= 1
    assert os.path.exists(SAMPLES), 'first create dataset'
    print 'seq+gibbs start {} passes'.format(passes)
    model = ImageModel()
    mixture = ImageModel.Mixture()
    mixture.init(model)
    scores = numpy.zeros(1, dtype=numpy.float32)
    assignments = {}

    for i, xy in enumerate(json_stream_load(SAMPLES)):
        scores.resize(len(mixture))
        mixture.score_value(model, xy, scores)
        groupid = sample_discrete_log(scores)
        mixture.add_value(model, groupid, xy)
        assignments[i] = mixture.id_tracker.packed_to_global(groupid)

    print 'seq+gibbs init with {} components'.format(len(mixture))

    for _ in xrange(passes - 1):
        for i, xy in enumerate(json_stream_load(SAMPLES)):
            groupid = mixture.id_tracker.global_to_packed(assignments[i])
            mixture.remove_value(model, groupid, xy)
            scores.resize(len(mixture))
            mixture.score_value(model, xy, scores)
            groupid = sample_discrete_log(scores)
            mixture.add_value(model, groupid, xy)
            assignments[i] = mixture.id_tracker.packed_to_global(groupid)

    print 'seq+gibbs found {} components'.format(len(mixture))
    image = synthesize_image(model, mixture)
    scipy.misc.imsave(os.path.join(RESULTS, 'seq_gibbs.png'), image)


def json_loop_load(filename):
    while True:
        for i, item in enumerate(json_stream_load(filename)):
            yield i, item


def annealing_schedule(passes):
    passes = float(passes)
    assert passes >= 1
    add_rate = passes
    remove_rate = passes - 1
    state = add_rate
    while True:
        if state >= 0:
            state -= remove_rate
            yield True
        else:
            state += add_rate
            yield False


@parsable.command
def compress_annealing(passes=PASSES):
    '''
    Compress image via subsample annealing.
    '''
    assert passes >= 1
    assert os.path.exists(SAMPLES), 'first create dataset'
    print 'annealing start {} passes'.format(passes)
    model = ImageModel()
    mixture = ImageModel.Mixture()
    mixture.init(model)
    scores = numpy.zeros(1, dtype=numpy.float32)
    assignments = {}

    to_add = json_loop_load(SAMPLES)
    to_remove = json_loop_load(SAMPLES)

    for next_action_is_add in annealing_schedule(passes):
        if next_action_is_add:
            i, xy = to_add.next()
            if i in assignments:
                break
            scores.resize(len(mixture))
            mixture.score_value(model, xy, scores)
            groupid = sample_discrete_log(scores)
            mixture.add_value(model, groupid, xy)
            assignments[i] = mixture.id_tracker.packed_to_global(groupid)
        else:
            i, xy = to_remove.next()
            groupid = mixture.id_tracker.global_to_packed(assignments.pop(i))
            mixture.remove_value(model, groupid, xy)

    print 'annealing found {} components'.format(len(mixture))
    image = synthesize_image(model, mixture)
    scipy.misc.imsave(os.path.join(RESULTS, 'annealing.png'), image)


@parsable.command
def clean():
    '''
    Clean out dataset and results.
    '''
    for dirname in [DATA, RESULTS]:
        if not os.path.exists(dirname):
            shutil.rmtree(dirname)


@parsable.command
def run(sample_count=SAMPLE_COUNT, passes=PASSES):
    '''
    Generate all datasets and run all algorithms.
    See index.html for results.
    '''
    create_dataset(sample_count)

    procs = [
        Process(target=compress_sequential),
        Process(target=compress_gibbs, args=(passes,)),
        Process(target=compress_annealing, args=(passes,)),
        Process(target=compress_seq_gibbs, args=(passes,)),
    ]
    for proc in procs:
        proc.start()
    for proc in procs:
        proc.join()

if __name__ == '__main__':
    parsable.dispatch()



# Copyright (c) 2014, Salesforce.com, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# - Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# - Neither the name of Salesforce.com nor the names of its contributors
#   may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy
import scipy.stats
from collections import defaultdict


def scores_to_probs(scores):
    scores = numpy.array(scores)
    scores -= scores.max()
    probs = numpy.exp(scores, out=scores)
    probs /= probs.sum()
    return probs


def score_to_empirical_kl(score, count):
    """
    Convert total log score to KL( empirical || model ),
    where the empirical pdf is uniform over `count` datapoints.
    """
    count = float(count)
    return -score / count - numpy.log(count)


def print_histogram(probs, counts):
    WIDTH = 60.0
    max_count = max(counts)
    print '{: >8} {: >8}'.format('Prob', 'Count')
    for prob, count in sorted(zip(probs, counts), reverse=True):
        width = int(round(WIDTH * count / max_count))
        print '{: >8.3f} {: >8d} {}'.format(prob, count, '-' * width)


def multinomial_goodness_of_fit(
        probs,
        counts,
        total_count,
        truncated=False,
        plot=False):
    """
    Pearson's chi^2 test, on possibly truncated data.
    http://en.wikipedia.org/wiki/Pearson%27s_chi-squared_test

    Returns:
        p-value of truncated multinomial sample.
    """
    assert len(probs) == len(counts)
    assert truncated or total_count == sum(counts)
    chi_squared = 0
    dof = 0
    if plot:
        print_histogram(probs, counts)
    for p, c in zip(probs, counts):
        if p == 1:
            return 1 if c == total_count else 0
        assert p < 1, 'bad probability: %g' % p
        if p > 0:
            mean = total_count * p
            variance = total_count * p * (1 - p)
            assert variance > 1,\
                'WARNING goodness of fit is inaccurate; use more samples'
            chi_squared += (c - mean) ** 2 / variance
            dof += 1
        else:
            print 'WARNING zero probability in goodness-of-fit test'
            if c > 0:
                return float('inf')

    if not truncated:
        dof -= 1

    survival = scipy.stats.chi2.sf(chi_squared, dof)
    return survival


def unif01_goodness_of_fit(samples, plot=False):
    """
    Bin uniformly distributed samples and apply Pearson's chi^2 test.
    """
    samples = numpy.array(samples, dtype=float)
    assert samples.min() >= 0.0
    assert samples.max() <= 1.0
    bin_count = int(round(len(samples) ** 0.333))
    assert bin_count >= 7, 'WARNING imprecise test, use more samples'
    probs = numpy.ones(bin_count, dtype=numpy.float) / bin_count
    counts = numpy.zeros(bin_count, dtype=numpy.int)
    for sample in samples:
        counts[int(bin_count * sample)] += 1
    return multinomial_goodness_of_fit(probs, counts, len(samples), plot=plot)


def density_goodness_of_fit(samples, probs, plot=False):
    """
    Transform arbitrary continuous samples to unif01 distribution
    and assess goodness of fit via Pearson's chi^2 test.

    Inputs:
        samples - a list of real-valued samples from a distribution
        probs - a list of probability densities evaluated at those samples
    """
    assert len(samples) == len(probs)
    assert len(samples) > 100, 'WARNING imprecision; use more samples'
    pairs = zip(samples, probs)
    pairs.sort()
    samples = numpy.array([x for x, p in pairs])
    probs = numpy.array([p for x, p in pairs])
    density = numpy.sqrt(probs[1:] * probs[:-1])
    gaps = samples[1:] - samples[:-1]
    unif01_samples = 1.0 - numpy.exp(-len(samples) * gaps * density)
    return unif01_goodness_of_fit(unif01_samples, plot=plot)


def discrete_goodness_of_fit(
        samples,
        probs_dict,
        truncate_beyond=8,
        plot=False):
    """
    Transform arbitrary discrete data to multinomial
    and assess goodness of fit via Pearson's chi^2 test.
    """
    assert len(samples) > 100, 'WARNING imprecision; use more samples'
    counts = defaultdict(lambda: 0)
    for sample in samples:
        assert sample in probs_dict
        counts[sample] += 1
    items = [(prob, counts.get(i, 0)) for i, prob in probs_dict.iteritems()]
    items.sort(reverse=True)
    truncated = (truncate_beyond and truncate_beyond < len(items))
    if truncated:
        items = items[:truncate_beyond]
    probs = [prob for prob, count in items]
    counts = [count for prob, count in items]
    return multinomial_goodness_of_fit(
        probs,
        counts,
        len(samples),
        truncated=truncated,
        plot=plot)


def bin_samples(samples, k=10, support=[]):
    """
    Bins a collection of univariate samples into k bins of equal
    fill via the empirical cdf, to be used in goodness of fit testing.

    Returns
    counts : array k x 1
    bin_ranges : arrary k x 2

    each count is the number of samples in [bin_min, bin_max)
    except for the last bin which is [bin_min, bin_max]

    list partitioning algorithm adapted from Mark Dickinson:
    http://stackoverflow.com/questions/2659900
    """
    samples = sorted(samples)

    N = len(samples)
    q, r = divmod(N, k)
    #we need to distribute the remainder relatively evenly
    #tests will be inaccurate if we have small bins at the end
    indices = [i * q + min(r, i) for i in range(k + 1)]
    bins = [samples[indices[i]: indices[i + 1]] for i in range(k)]
    bin_ranges = []
    counts = []
    for i in range(k):
        bin_min = bins[i][0]
        try:
            bin_max = bins[i + 1][0]
        except IndexError:
            bin_max = bins[i][-1]
        bin_ranges.append([bin_min, bin_max])
        counts.append(len(bins[i]))
    if support:
        bin_ranges[0][0] = support[0]
        bin_ranges[-1][1] = support[1]
    return numpy.array(counts), numpy.array(bin_ranges)


def histogram(samples, bin_count=None):
    if bin_count is None:
        bin_count = numpy.max(samples) + 1
    v = numpy.zeros(bin_count, dtype=int)
    for sample in samples:
        v[sample] += 1
    return v
